from Bubot.Helpers.Helper import Helper
from Bubot.Helpers.ExtException import ExtException, HandlerNotFoundError


class BubotHelper:

    @staticmethod
    def get_obj_class(obj_name, **kwargs):
        suffix = kwargs.get('suffix')
        class_name = f'{obj_name}{suffix}' if suffix else obj_name
        full_path = f'BubotObj.{obj_name}.{class_name}.{class_name}'
        try:
            return Helper.get_class(full_path)
        except ExtException as err:
            raise HandlerNotFoundError(detail=f'object {obj_name}', parent=err)

    @staticmethod
    def get_subtype_class(obj_name, subtype, **kwargs):
        suffix = kwargs.get('suffix')
        class_name = f'{subtype}{suffix}' if suffix else subtype
        full_path = f'BubotObj.{obj_name}.subtype.{subtype}.{class_name}.{class_name}'
        try:
            return Helper.get_class(full_path)
        except ExtException as err:
            raise HandlerNotFoundError(detail=f'object {obj_name} subtype {subtype}', parent=err)

    @staticmethod
    def get_extension_class(obj_name, device, **kwargs):
        suffix = kwargs.get('suffix')
        class_name = f'{obj_name}{suffix}' if suffix else obj_name
        full_path = f'BubotObj.{obj_name}.extension.{device}.{class_name}.{class_name}'
        try:
            return Helper.get_class(full_path)
        except ExtException as err:
            raise HandlerNotFoundError(detail=f'object {obj_name} extension {device}', parent=err)



from BubotObj.OcfDevice.subtype.Device.Device import Device
from .__init__ import __version__ as device_version
import asyncio


# _logger = logging.getLogger(__name__)


class EchoDevice(Device):
    version = device_version
    template = False
    file = __file__

    async def on_idle(self):
        i = self.get_param('/oic/mnt', 'i')
        self.set_param('/oic/mnt', 'i', i + 1)
        self.log.info('i: {}'.format(i))
        await asyncio.sleep(1)

    @classmethod
    def get_install_actions(cls):
        result = [
            cls.get_install_search_action()
        ]
        result.extend(super().get_install_actions())
        return result

    async def on_action(self, message, answer):
        j = self.get_param('/oic/mnt', 'j')
        self.set_param('/oic/mnt', 'j', j + 1)
        self.log.info('j: {}'.format(j))

    async def find_devices(self, **kwargs):
        notify = kwargs.get('notify')
        if notify:
            await notify({'message': f'Ищем {self.__class__.__name__}...'})
        result = []
        for i in range(2):
            await asyncio.sleep(1)
            if notify:
                await notify({'message': f'Найдено {self.__class__.__name__}: {i + 1}'})
            tmp_device = EchoDevice.init_from_config({})
            result.append(dict(
                id=tmp_device.get_device_id(),
                name=tmp_device.get_device_name(),
                links=tmp_device.get_discover_res(),
                _actions=Device.get_install_actions()
            ))
        return result



"""
TODO Проверка смены IP адреса и автоматическая замена на актуальный

"""

from .MainLoopMixin import MainLoopMixin
from Bubot.Helpers.ExtException import ExtException, ExtTimeoutError
# from .QueueMixin import QueueMixin
from Bubot.Core.OcfMessage import OcfRequest
from Bubot.Helpers.Helper import Helper
from .__init__ import __version__ as device_version
from os import path
import os
import json
import asyncio
import logging
import re


# _logger = logging.getLogger('OcfDevice')


class Device(MainLoopMixin):
    scheme = {}
    cache = {}
    file = __file__
    version = device_version
    platform_version = device_version
    template = True

    def __init__(self, **kwargs):
        MainLoopMixin.__init__(self, **kwargs)
        self.loop = kwargs.get('loop', asyncio.get_event_loop())

    def run(self):
        self.loop = asyncio.get_event_loop()
        self.task = self.loop.create_task(self.main())
        self.loop.run_forever()

    @classmethod
    def find_first_config(cls, config_path, class_name):
        _list = os.listdir(config_path)
        pattern = re.compile('{0}.+.json'.format(class_name))
        for _file in _list:
            if os.path.isfile('{0}/{1}'.format(config_path, _file)):
                if pattern.match(_file):
                    return _file.split('.')[1]
        return None

    @classmethod
    def init_from_file(cls, **kwargs):
        kwargs['path'] = os.path.abspath(kwargs.get('path', './'))
        kwargs['log'] = kwargs['log'] if kwargs.get('log') else logging.getLogger()
        class_name = kwargs.get('class_name', cls.__name__)
        di = kwargs.get('di')
        config = {}
        if di is None:
            di = cls.find_first_config('{}'.format(kwargs['path']), class_name)
        if di:
            config_path = path.normpath('{0}/{1}.{2}.json'.format(kwargs['path'], class_name, di))
            try:
                with open(config_path, encoding='utf-8') as file:
                    config = json.load(file)
                    kwargs['log'].info('OcfDevice.init_from_file {0}.{1}'.format(class_name, di))
            except FileNotFoundError:
                kwargs['log'].warning('OcfDevice config not found {0}'.format(config_path))
            except Exception as e:
                raise ExtException(
                    7004,
                    detail='{0} {1}'.format(str(e), config_path),
                    action='OcfDevice.init_from_config',
                    dump=dict(
                        class_name=class_name,
                        di=di,
                    )
                )
        kwargs['class_name'] = class_name
        kwargs['di'] = di
        return cls.init_from_config(config, **kwargs)

    @classmethod
    def init_from_config(cls, config=None, **kwargs):
        cache = kwargs.get('cache', cls.cache)
        class_name = cls.__name__
        try:
            if config:
                class_name = config['/oic/d']['dmno']
        except KeyError:
            pass
        class_name = kwargs.get('class_name', class_name)
        try:
            self = cls.get_device_class(class_name)
            self = self(**kwargs)
        except Exception as err:
            raise ExtException(
                'Bad driver',
                detail=class_name,
                action='OcfDevice.init_from_config',
                parent=err,
                dump=dict(
                    config=config,
                    kwargs=kwargs
                )
            )
        try:
            self.data = self.get_default_config(self.__class__, Device, cache)
            if config:
                Helper.update_dict(self.data, config)
            if not self.get_device_id():
                self.set_device_id(kwargs.get('di'))

            # self.log = Logger(
            #     self,
            #     level=self.get_param('/oic/con', 'logLevel', 'error'),
            #     name='{0}:{1}'.format(self.__class__.__name__, self.get_device_id()[-5:]),
            #     log=kwargs.get('log')
            # )
            self.log = logging.getLogger('{0}:{1}'.format(self.__class__.__name__, self.get_device_id()[-5:]))
            self.log.setLevel(getattr(logging, self.get_param('/oic/con', 'logLevel', 'error').upper()))
            return self
        except Exception as e:
            raise ExtException(
                'Bad driver config',
                detail=class_name,
                action='OcfDevice.init_from_config',
                dump=dict(
                    config=config,
                    kwargs=kwargs
                )
            ) from e

    def get_default_config(self, current_class, root_class, cache):
        data = Helper.get_default_config(current_class, root_class, cache)
        data['/oic/d']['dmno'] = current_class.__name__
        data['/oic/d']['sv'] = self.version
        data['/oic/p']['mnpv'] = current_class.platform_version
        return data

    def save_config(self):
        def_data = self.get_default_config(self.__class__, Device, self.cache)
        data = Helper.compare(def_data, self.data)
        try:
            data[1].pop('/oic/mnt')
        except KeyError:
            pass

        try:
            with open(self.get_config_path(), 'w', encoding='utf-8') as file:
                json.dump(data[1], file, ensure_ascii=False, indent=2)
        except FileNotFoundError:
            return {}
        return data

    async def request(self, operation, to, data=None, **kwargs):
        try:
            msg = OcfRequest(
                to=to,
                fr=self.link,
                op=operation,
                cn=data,
                # uri_path=link['href'],
                # operation=operation,
                # data=data,
                # code=kwargs.pop('code', 1),
                token=self.coap.token,
                mid=self.coap.mid,
                **kwargs
            )
            coap_msg, remote = msg.encode_to_coap()
            result = await self.coap.send_request(coap_msg, remote)
            return result
        except TimeoutError:
            raise ExtTimeoutError(action='request', dump=dict(op=operation, to=to, data=data, kwargs=kwargs)) from None
        except ExtException as err:
            raise ExtException(parent=err,
                               action='{}.request()'.format(self.__class__.__name__),
                               dump=dict(op=operation, to=to, data=data, kwargs=kwargs)) from None
        except Exception as err:
            raise ExtException(parent=err,
                               action='{}.request()'.format(self.__class__.__name__),
                               dump=dict(op=operation, to=to, data=data, kwargs=kwargs)) from None

    async def observe(self, to, callback=None):
        try:
            token = self.coap.token
            msg = OcfRequest(
                to=to,
                fr=self.link,
                op='retrieve',
                token=token,
                mid=self.coap.mid,
                obs=1 if callback is None else 0
            )
            coap_msg, remote = msg.encode_to_coap()
            await self.coap.send_multi_answer_request(coap_msg, remote, callback)
            if callback is None:
                del self.coap.answer[token]
        except TimeoutError as e:
            raise ExtTimeoutError(action='request',
                                  dump=dict(op='observe', to=to)) from None
        except ExtException as e:
            raise ExtException(parent=e,
                               action='{}.request()'.format(self.__class__.__name__),
                               dump=dict(op='observe', to=to)) from None
        except Exception as e:
            raise ExtException(parent=e,
                               action='{}.request()'.format(self.__class__.__name__),
                               dump=dict(op='observe', to=to)) from None

    async def discovery_resource(self, **kwargs):
        try:
            token = self.coap.token
            result = {}
            msg = OcfRequest(
                to=dict(href='/oic/res'),
                fr=self.link,
                op='retrieve',
                token=token,
                mid=self.coap.mid,
                multicast=True,
                **kwargs
            )
            coap_msg, remote = msg.encode_to_coap()
            if self.coap.ipv6:
                await self.coap.send_multi_answer_request(
                    coap_msg,
                    (self.coap.coap_discovery_ipv6[0], self.coap.multicast_port),
                    self.on_response_oic_res,
                    result
                )
            if self.coap.ipv4:
                await self.coap.send_multi_answer_request(
                    coap_msg,
                    (self.coap.coap_discovery_ipv4[0], self.coap.multicast_port),
                    self.on_response_oic_res,
                    result
                )
            await asyncio.sleep(2)
            result = self.coap.answer[token]['result']
            # del (self.coap.answer[token])
            return result

        except ExtException as e:
            raise Exception(e)
        except Exception as e:
            raise ExtException(e)

    async def find_resource_by_link(self, link):
        self.log.debug('find resource by di {0} href {1}'.format(link.di, link.href))

        links = await self.discovery_resource(
            query=dict(di=[link.di], href=[link.href]))
        if isinstance(links, dict):
            for di in links:
                if di == link.di:
                    for ref in links[di].links:
                        if ref == link.href:
                            return links[di].links[ref]
        return None



from Bubot.Core.ObjForm import ObjForm
import asyncio
from Bubot.Helpers.Action import async_action
from BubotObj.OcfDevice.subtype.WebServer.ApiHelper import DeviceApi


class CatalogObjApi(DeviceApi):
    handler = None
    extension = False

    @async_action
    async def api_read(self, view, **kwargs):
        # action = kwargs['_action']
        org = self.handler(view.storage, account_id=view.session.get('account', 'Bubot'))
        _id = view.request.query.get('id')
        result = await org.find_by_id(_id)
        return self.response.json_response(result)

    @async_action
    async def api_create(self, view, **kwargs):
        handler, data = await self.prepare_json_request(view)
        handler.init_by_data(data)
        await handler.update()
        await handler.update()
        return self.response.json_response(handler.data)

    @async_action
    async def api_delete(self, view, **kwargs):
        handler, data = await self.prepare_json_request(view)
        await handler.delete_one(data['_id'])
        await handler.update()
        return self.response.json_response(handler.data)

    @async_action
    async def api_update(self, view, **kwargs):
        handler, data = await self.prepare_json_request(view)
        handler.init_by_data(data)
        await handler.update()
        return self.response.json_response(handler.data)

    @async_action
    async def api_query(self, view, **kwargs):
        handler, data = await self.prepare_json_request(view, **kwargs)
        # file_name = '{}/examples/test-query-response.json'.format(os.path.dirname(__file__))
        # with open(file_name, 'r', encoding='utf-8') as file:
        #     data = json.load(file)
        # obj = self.handler(view.storage, account_id=view.session['account'])
        filter = self.prepare_query_filter(data)
        data = await handler.query(**filter)

        # await asyncio.sleep(2)
        return self.response.json_response({"rows": data})

    async def prepare_json_request(self, view, **kwargs):
        handler = self.handler(view.storage, account_id=view.session.get('account'))
        handler.init()
        data = await view.loads_json_request_data(view)
        return handler, data

    def prepare_query_filter(self, data):
        page = data.pop('page', None)
        filter = {}
        limit = min(self.query_limit, int(data.pop('limit', self.query_limit)))
        for key in data:
            try:
                filter[key] = self.filter_fields[key](filter, key, data[key])
            except:
                filter[key] = data[key]
        result = {
            'filter': filter
        }

        if limit:
            result['limit'] = limit
            if page:
                result['skip'] = (int(page) - 1) * result['limit']
        return result



import asyncio
import time
from Bubot.Helpers.Coap.CoapServer import CoapServer
from BubotObj.OcfDevice.subtype.Device.DeviceCore import DeviceCore
from Bubot.Helpers.ExtException import ExtException
import random


class MainLoopMixin(DeviceCore):
    # _logger = logging.getLogger(__name__)

    async def main(self):
        try:
            self.log.info("begin main")
            self.coap = CoapServer(self)
            await self.coap.run()

            update_time = self.get_param('/oic/con', 'updateTime', 30)
            await asyncio.sleep(random.random())

            while True:
                last_run = time.time()
                self.set_param('/oic/mnt', 'lastRun', last_run)
                status = self.get_param('/oic/mnt', 'currentMachineState', 'pending')
                method = f'on_{status}'
                try:
                    await getattr(self, method)()
                    await asyncio.sleep(0.01)
                except asyncio.CancelledError:
                    if self.get_param('/oic/mnt', 'currentMachineState', 'pending'):
                        self.set_param('/oic/mnt', 'currentMachineState', 'cancelled')
                    else:
                        self.log.info("cancelled main")
                        return
                        # raise asyncio.CancelledError
                except ExtException as err:
                    self.log.error(ExtException(detail=method, parent=err))
                    self.set_param('/oic/mnt', 'currentMachineState', 'stopped')
                    self.set_param('/oic/mnt', 'message', str(err))
                    pass
                except Exception as err:
                    self.log.error(ExtException(detail=method, parent=err))
                    self.set_param('/oic/mnt', 'currentMachineState', 'stopped')
                    self.set_param('/oic/mnt', 'message', str(err))
                    pass

                await self.check_changes()
                elapsed_time = time.time() - last_run
                sleep_time = max(0.01, max(update_time - elapsed_time, 0))
                await asyncio.sleep(sleep_time)
        except ExtException as err:
            self.log.error(f"end main {err}")
            raise
        pass

    async def on_pending(self):
        self.set_param('/oic/mnt', 'currentMachineState', 'idle')

    async def on_idle(self):
        pass

    async def on_cancelled(self):
        self.coap.close()
        self.set_param('/oic/mnt', 'currentMachineState', '')
        raise asyncio.CancelledError()

    async def on_stopped(self):
        try:
            await asyncio.Future()
        except asyncio.CancelledError:
            self.set_param('/oic/mnt', 'status', 'cancelled')



from Bubot.Helpers.Helper import Helper
from Bubot.Helpers.ExtException import ExtException, KeyNotFound, ExtTimeoutError
import asyncio
from Bubot.Core.OcfMessage import OcfResponse, OcfRequest
from Bubot.Core.BubotHelper import BubotHelper
from uuid import uuid4
import logging
import random
from Bubot.Core.DeviceLink import DeviceLink, ResourceLink
import os.path


# self.logger = logging.getLogger('DeviceCore')


class DeviceCore:

    def __init__(self, **kwargs):
        self.log = None
        self.resource = {}
        self.data = {}
        self.coap = None
        self._resource_changed = {}
        self._link = None
        self.loop = None
        self.task = None
        self.path = os.path.abspath(kwargs.get('path', './'))

    def get_device_id(self):
        return self.data['/oic/d'].get('di')
        # return self.data['/oic/d'].get('piid')

    def get_device_name(self):
        try:
            return self.data['/oic/d'].get('n')
        except Exception:
            return self.__class__.__name__

    def set_device_id(self, device_id=None):
        if device_id is None:
            device_id = str(uuid4())
        self.data['/oic/d']['di'] = device_id
        # self.data['/oic/d']['piid'] = device_id

    def get_param(self, uri, *args):
        try:
            _res = self.data[uri]
        except KeyError:
            raise KeyNotFound(
                action='OcfDevice.get_param',
                detail=f'{uri} ({self.__class__.__name__})'
            ) from None
        if not args or args[0] is None:
            return _res
        try:
            return _res[args[0]]
        except KeyError:
            try:
                return args[1]
            except IndexError:
                raise KeyNotFound(
                    action='OcfDevice.get_param',
                    detail=f'{args[0]} ({self.__class__.__name__}{uri})'
                ) from None

    def set_param(self, resource, name, new_value, **kwargs):
        try:
            old_value = self.get_param(resource, name, new_value)
            difference, changes = Helper.compare(old_value, new_value)
            if difference:
                self._resource_changed[resource] = True
                self.data[resource][name] = new_value

                if kwargs.get('save_config', False):
                    self.save_config()
        except Exception as e:
            raise ExtException(detail=str(e), dump=dict(
                resource=resource,
                name=name,
                value=new_value
            ))

    def update_param(self, resource, name, new_value, **kwargs):
        old_value = self.get_param(resource, name)
        difference, changes = Helper.compare(old_value, new_value)
        if difference:
            self._resource_changed[resource] = True
            if isinstance(old_value, dict):
                Helper.update_dict(old_value, changes)
            elif isinstance(old_value, (str, int, float, bool)):
                self.data[resource][name] = changes
            elif isinstance(old_value, list):
                self.log.warning("NOT SUPPORTED OPERATIONS!!!")
                self.data[resource][name] = changes

            if kwargs.get('save_config', False):
                self.save_config()
        return changes

    def on_update_oic_con(self, message):
        changes = self.update_param(message.to.href, None, message.cn)
        if 'log_level' in changes:
            self.log.setLevel(getattr(logging, self.get_param('/oic/con', 'logLevel', 'error').upper()))
        return changes

    def get_config_path(self):
        return os.path.join(self.path, f'{self.__class__.__name__}.{self.get_device_id()}.json')

    def delete_config(self):
        os.remove(self.get_config_path())

    def save_config(self):
        raise NotImplemented()
        # listener = self.listener.get(name, [])
        # task = []
        # for device in listener:
        #     task.append(self.send_event_change(name, device))
        # await asyncio.gather(task)

    async def send_event_change(self, resource_name, receiver):
        pass

    @property
    def link(self):
        return self.get_link()

    def get_link(self, href=None):
        if self._link is None:
            eps = []
            if not self.coap:
                raise KeyNotFound(detail='COAP socket not found')
            for elem in self.coap.endpoint:
                if elem == 'multicast' or not self.coap.endpoint[elem]:
                    continue
                eps.append(dict(ep=self.coap.endpoint[elem]['uri']))
            self._link = dict(
                anchor='ocf://{}'.format(self.get_device_id()),
                eps=eps
            )
        if href and href in self.data:
            return dict(
                anchor=self._link['anchor'],
                eps=self._link['eps'],
                href=href

            )
        else:
            return self._link

    async def on_get_request(self, message):
        # interface = message.uri_query['if'][0] if 'if' in message.uri_query else 'oic.if.baseline'
        props = 'on_{0}{1}'.format(message.op, message.to.href.replace('/', '_'))
        if message.obs is not None:  # observe request
            if message.obs == 1:  # cancel observer
                self.del_listener(message.to.href, message.fr.uid)
            elif message.obs == 0:  # observe request
                self.add_listener(message.to.href, message.fr, message.token)
                pass
        if hasattr(self, props):
            return await getattr(self, props)(message)
        return self.get_param(message.to.href)

    async def on_post_request(self, message):
        # interface = message.uri_query['if'][0] if 'if' in message.uri_query else 'oic.if.baseline'
        props = 'on_{0}{1}'.format(message.op, message.to.href.replace('/', '_'))
        if hasattr(self, props):
            logging.debug(props)
            return await getattr(self, props)(message)
        logging.debug('on_post_request')
        return self.update_param(message.to.href, None, message.cn)

    async def on_retrieve_oic_res(self, message):
        device = DeviceLink.init_from_device(self)
        links = []
        for link in device.links:
            if device.links[link].discoverable:
                suited = True
                if message.query:
                    for key in message.query:
                        if not suited:
                            break
                        try:
                            value = getattr(device.links[link], key)
                            if isinstance(value, list):
                                for elem in message.query[key]:
                                    if elem not in value:
                                        suited = False
                                        break
                            elif value not in message.query[key]:
                                suited = False
                        except AttributeError:
                            suited = False
                if suited:
                    links.append(device.links[link].data)
        self.log.debug('discovery {0} links'.format(len(links)))
        await asyncio.sleep(random.random())
        if links:
            return [{
                'di': device.di,
                'n': device.name,
                'links': links
            }]
        raise asyncio.CancelledError

    async def on_response_oic_res(self, message, answer):
        # self.log.debug('begin from {}'.format(message.to.uid))
        async with answer['lock']:
            if answer['result'] is None:
                answer['result'] = {}
            if message.is_successful():
                if message.cn:
                    for device in message.cn:
                        if device['di'] in answer['result']:
                            answer['result'][device['di']].update_from_oic_res(device)
                        else:
                            answer['result'][device['di']] = DeviceLink.init_from_oic_res(device)
            else:
                self.log.error('{0}'.format(message.cn))
        # self.log.debug('end from {}'.format(message.to.uid))

    async def on_update_oic_mnt(self, message):
        result = self.update_param(message.to.href, None, message.cn)
        state = message.cn.get('currentMachineState')
        if state:
            if state == 'cancelled':
                self.loop.create_task(self.cancel())

    async def cancel(self):
        self.log.debug('begin cancelled')
        self.set_param('/oic/mnt', 'currentMachineState', 'cancelled')
        try:
            # await self.task
            await asyncio.wait_for(self.task, 30)
        except asyncio.TimeoutError:
            self.task.cancel()

    def get_discover_res(self):
        result = {}
        for href in self.data:
            try:
                if self.data[href]['p']['bm'] == 0:
                    continue
                result[href] = self.data[href]
            except:
                result[href] = self.data[href]
        return result

    @classmethod
    def get_install_search_action(cls):
        return dict(
            name='CallDataSourceForSelectedItems',
            icon='mdi-radar',
            title='search OcfDriver',
            data={
                "method": "find_devices",
                "operation": {
                    "title": "Find devices",
                    "cancelable": True,
                    "show": True,
                    "formUid": "OcfDriver/FoundDevices"
                },
                "dataSource": {
                    "type": "Vuex",
                    "storeName": "LongOperations",
                    "dispatchName": "run",
                    "keyProperty": "id",
                    "objName": "OcfDriver"
                }
            }
        )

    @classmethod
    def get_install_actions(cls):
        return [
            # dict(
            #     name='CallDataSourceForSelectedItems',
            #     icon='mdi-plus-circle-outline',
            #     title='add devices',
            #     data=dict(
            #         method='add_devices'
            #     )
            # )
        ]

    def add_listener(self, href, link, token):
        listening = self.get_param('/oic/con', 'listening')
        for elem in listening:
            if elem['href'] == href and elem['di'] == link['di']:
                return
        listening.append(dict(
            href=href,
            di=link.di,
            ep=link.get_endpoint(),
            token=token
        ))
        self.set_param('/oic/con', 'listening', listening)

    def del_listener(self, href, di):
        listening = self.get_param('/oic/con', 'listening')
        change = False
        for i, elem in enumerate(listening):
            if elem['href'] == href and elem['di'] == di:
                del listening[i]
                change = True
                break
        if change:
            self.set_param('/oic/con', 'listening', listening)

    async def check_changes(self):
        if self._resource_changed:
            for href in self._resource_changed:
                if self._resource_changed[href]:
                    self._resource_changed[href] = False
                    await self.notify(href, self.get_param(href))

    async def notify(self, href, data):  # send notify response to observer
        listening = self.get_param('/oic/con', 'listening')
        if not listening:
            return
        to = ResourceLink.init_from_link(self.link)
        to.href = href
        for elem in listening:
            if elem['href'] != href:
                continue
            try:
                self.log.debug('notify {0} {1}'.format(self.get_device_id(), href))
                msg = OcfResponse.generate_answer(data, OcfRequest(
                    to=to,
                    fr=ResourceLink.init_from_link(dict(di=elem['di'], ep=elem['ep'])),
                    op='retrieve',
                    token=elem['token'],
                    mid=self.coap.mid,
                    obs=0
                ))
                await self.coap.send_answer(msg)
            except TimeoutError as e:
                raise Exception(9001, action='notify',
                                dump=dict(op='observe', to=to)) from None
            except ExtException as e:
                raise ExtException(parent=e,
                                   action='{}.notify()'.format(self.__class__.__name__),
                                   dump=dict(op='observe', to=to)) from None
            except Exception as e:
                raise ExtException(parent=e,
                                   action='{}.notify()'.format(self.__class__.__name__),
                                   dump=dict(op='observe', to=to)) from None

    async def find_devices(self, **kwargs):
        raise NotImplemented()

    @classmethod
    def get_device_class(cls, class_name):
        return BubotHelper.get_subtype_class('OcfDevice', class_name, folder=True)



import multiprocessing
from Bubot.Core.DeviceLink import ResourceLink
import queue
from BubotObj.OcfDevice.subtype.Device.Device import Device
from BubotObj.OcfDevice.subtype.VirtualServer import __version__ as device_version
import logging
import logging.handlers
import asyncio
from Bubot.Helpers.ExtException import KeyNotFound
from Bubot.Helpers.Helper import ArrayHelper
import concurrent.futures


# _logger = multiprocessing.get_logger()


class VirtualServer(Device):
    version = device_version
    template = False
    file = __file__
    run_dev = ('/oic/con', 'running_devices')

    def __init__(self, **kwargs):
        self._running_devices = {}
        self.loop = None
        self.task = None
        self.queue = None
        Device.__init__(self, **kwargs)

    def run(self):
        self.log.debug('begin')
        if multiprocessing.current_process().name == 'MainProcess':
            self.queue = multiprocessing.Queue(-1)
            self.task = self.loop.create_task(self.logger())

        self.task = self.loop.create_task(self.main())

        if not self.loop.is_running():
            self.loop.run_forever()  # todo добавиь обработку Ctrl + C
        self.log.debug('end')
        pass

    async def logger(self):
        def _get(_queue):
            return _queue.get()

        executor = concurrent.futures.ThreadPoolExecutor()
        while True:
            try:
                record = await self.loop.run_in_executor(executor, _get, self.queue)
                if record is None:  # We send this as a sentinel to tell the listener to quit.
                    break
                logger = logging.getLogger(record.name)
                logger.handle(record)  # No level or filter logic applied - just do it!
            except queue.Empty:
                await asyncio.sleep(0.5)
            except Exception as e:
                import sys, traceback
                print('Whoops! Problem:', file=sys.stderr)
                traceback.print_exc(file=sys.stderr)

    async def on_pending(self):
        links = self.get_param('/oic/con', 'running_devices')
        if links:
            for link in links:
                device = await self.action_run_device(link)
                if device:
                    link['n'] = device.get_device_name()
        await super().on_pending()
        self.save_config()

    async def on_cancelled(self):
        for di in self._running_devices.keys():
            device = self._running_devices.pop(di)
            self.log.debug('{} begin cancelled'.format(di))
            if isinstance(device, multiprocessing.Process):
                res = await self.find_resource_by_link(ResourceLink.init_from_link(dict(di=di, href='/oic/mnt')))
                await self.request('update', res, dict(currentMachineState='cancelled'))
                for i in range(15):
                    if not device.is_alive():
                        break
                    self.log.debug('wait cancelled {}'.format(di))
                    await asyncio.sleep(i)
            else:
                await device.cancel()

            self.log.debug('{} end cancelled'.format(di))
        await super().on_cancelled()

    async def on_stopped(self):
        pass
        # links = self.get_param(*self.run_dev)
        # for i, link in enumerate(links):
        #     self._running_devices[link['di']][1].cancel()
        # await super().on_stopped()

    async def on_update_oic_con(self, message):
        # if 'running_devices' in result:
        #     new_links =

        # async def post_devices(self, message):

        links = self.get_param(*self.run_dev)
        new_links = message.cn.pop('running_devices', [])

        # index_current_link = Helper.index_list(current_link, 'di')
        index_list = ArrayHelper.index_list(new_links, 'di')
        changed_links = False
        for link in reversed(links):
            if link['di'] not in index_list:
                changed_links = True
                await self.action_del_device(link['di'])

        index_list = ArrayHelper.index_list(links, 'di')
        for link in reversed(new_links):
            if 'di' not in link or link['di'] not in index_list:
                changed_links = True
                await self.action_add_device(link)
        # изменять существующие записи нельзя
        self.set_param(*self.run_dev, links)
        result = self.update_param(message.to.href, None, message.cn)
        if changed_links:
            result['running_devices'] = self.get_param(*self.run_dev)
        return result

    async def action_add_device(self, link):
        await self.action_run_device(link)
        links = self.get_param(*self.run_dev)
        ArrayHelper.update(links, link, 'di')
        self.set_param(*self.run_dev, links)
        self.save_config()
        pass

    async def action_del_device(self, di):
        links = self.get_param(*self.run_dev)
        for i, link in enumerate(links):
            if link.get('di') == di:
                await self.action_stop_device(di)

                del links[i]
                self.set_param(*self.run_dev, links)
                self.save_config()
                return
        raise Exception('OcfDevice not found')

    async def action_run_device(self, link):
        di = link.get('di')
        try:
            class_name = link['dmno']
        except KeyError:
            raise KeyNotFound(detail='dmno', action='action_run_device') from None
        if di and di in self._running_devices:
            raise Exception('device is already running')
        if class_name == 'VirtualServer':
            process = multiprocessing.Process(
                target=self.device_process,
                args=(class_name, di, self.queue, dict(path=self.path)),
                daemon=True
            )
            process.start()
            self._running_devices[link['di']] = process
            return None

        else:
            device = Device.init_from_file(
                class_name=class_name,
                di=link.get('di'),
                path=self.path,
                loop=self.loop,
                log=self.log
            )
            link['di'] = device.get_device_id()
            task = self.loop.create_task(device.main())
            device.task = task
            self._running_devices[link['di']] = device
            return device

    async def action_stop_device(self, di):
        try:
            self._running_devices[di].task.cancel()
            await self._running_devices[di].task
        except asyncio.CancelledError:
            pass

    @staticmethod
    def device_process(class_name, di, queue, kwargs):
        h = logging.handlers.QueueHandler(queue)  # Just the one handler needed
        kwargs['loop'] = asyncio.new_event_loop()
        root = logging.getLogger()
        root.handlers = []
        root.addHandler(h)
        device = Device.init_from_file(
            class_name=class_name,
            di=di,
            **kwargs
        )
        device.run()



class PythonFastStorage:
    def __init__(self):
        self.storage = {}

    async def set(self, key, value, **kwargs):
        self.storage[key] = value

    async def get(self, key, **kwargs):
        return self.storage[key]



from os.path import dirname, normpath, isdir, isfile
from json import load


class ObjModel:
    def __init__(self, **kwargs):
        self.data = kwargs.get('data')

    @classmethod
    def get(cls, obj_handler):
        form_path = normpath(f'{dirname(obj_handler.file)}/{obj_handler.get_name}.model.json')
        if not isfile(form_path):
            return None
        with open(form_path, 'r', encoding='utf-8') as file:
            return cls(data=load(file))

    def get_projections(self):
        fields = self.data.get('fields')
        if not fields:
            return {'_id': 1, 'title': 1}
        res = {}
        for elem in fields:
            if fields[elem]:
                res[elem]: 1
        return res



from aiohttp import web
from Bubot.Core.FastStorage.PythonFastStorage import PythonFastStorage as FastStorage
from Bubot.Core.DataBase.Mongo import Mongo as Storage
from Bubot.Core.AuthHandler import AuthHandler
from Bubot.Core.ApiHandler import ApiHandler
from Bubot.Core.FormHandler import FormHandler
from Bubot.Core.ReportHandler import ReportHandler


def start():
    app = web.Application()
    # logging.basicConfig(level=logging.DEBUG)
    app.router.add_route('*', '/auth/login_by_cert', AuthHandler)
    app.router.add_route('*', '/api/{objType}/{objName}', ApiHandler)
    app.router.add_route('*', '/api/{objType}/{objName}/{method}', ApiHandler)
    app.router.add_route('*', '/form/{objType}/{objName}/{formName}', FormHandler)
    app.router.add_route('*', '/report/{objType}/{objName}/{reportName}/{reportSection}', ReportHandler)
    app.router.add_static('/jr', './jasper_reports')
    app['fast_storage'] = FastStorage()
    app['storage'] = Storage.connect()
    web.run_app(app, port=8081)


if __name__ == '__main__':
    start()



from BubotObj.OcfDevice.subtype.Device.Device import Device
from Bubot.Helpers.ExtException import ExtException
from sys import path as syspath
import os


def find_drivers(**kwargs):
    result = {}
    log = kwargs.get('log')
    for path1 in syspath:
        device_dir = os.path.normpath('{}/BubotObj/OcfDevice/subtype'.format(path1))
        if os.path.isdir(device_dir):
            device_list = os.listdir(device_dir)
            for device_name in device_list:
                device_path = os.path.normpath('{0}/{1}/{1}.py'.format(device_dir, device_name))
                if not os.path.isfile(device_path):
                    continue
                try:
                    driver = Device.init_from_config(class_name=device_name)
                except ExtException as err:
                    if log:
                        log.error(err)
                    continue
                if driver.template:
                    continue
                add_driver = True
                filter_rt = kwargs.get('rt')
                if filter_rt:
                    try:
                        find = False
                        for href in driver.data:
                            rt = driver.data[href].get('rt')
                            if filter_rt in rt:
                                find = True
                                break
                        if not find:
                            add_driver = False
                    except Exception as e:
                        add_driver = False
                if add_driver:
                    result[device_name] = dict(
                        path=os.path.normpath('{0}/{1}'.format(device_dir, device_name)),
                        # driver=driver
                    )

    return result


def find_schemas(**kwargs):
    result = []
    for path1 in syspath:
        schemas_dir = os.path.normpath('{}/BubotObj/OcfSchema/schema'.format(path1))
        if os.path.isdir(schemas_dir) and schemas_dir not in result:
            result.append(schemas_dir)

    return result




import asyncio
import inspect
from os import path


async def wait_run_device(device):
    device_task = device.loop.create_task(device.main())
    while device.get_param('/oic/mnt', 'currentMachineState') != 'idle':
        try:
            device_task.result()
        except asyncio.InvalidStateError:
            pass
        await asyncio.sleep(0.1)
    return device_task


async def wait_run_device2(device):
    while device.get_param('/oic/mnt', 'currentMachineState') != 'idle':
        try:
            device.task.result()
        except asyncio.InvalidStateError:
            pass
        await asyncio.sleep(0.1)
    return device.task


async def wait_cancelled_device(device, device_task):
    try:
        device.set_param('/oic/mnt', 'currentMachineState', 'cancelled')
        while device.get_param('/oic/mnt', 'currentMachineState') != '':
            try:
                device_task.result()
            except asyncio.InvalidStateError:
                pass
            await asyncio.sleep(0.1)
    except asyncio.CancelledError:
        pass
    await device_task
    pass


def async_test(f):
    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        if inspect.iscoroutinefunction(f):
            future = f(*args)
        else:
            coroutine = asyncio.coroutine(f)
            future = coroutine(*args, **kwargs)
        loop.run_until_complete(future)

    return wrapper


def get_config_path(_file_):
    return '{}/config/'.format(path.dirname(_file_))



from motor import motor_asyncio
from urllib.parse import quote_plus
from Bubot.Helpers.Action import async_action


class Mongo:
    def __init__(self, **kwargs):
        self.client = kwargs.get('client')

    pass

    @classmethod
    def connect(cls, device, **kwargs):
        user = kwargs.get('user')
        if user:
            uri = "mongodb://{user}:{password}@{host}:{port}".format(
                user=quote_plus(user),
                password=quote_plus(kwargs.get('password')),
                host=kwargs.get('host', 'localhost'),
                port=kwargs.get('port', 27017)
            )
        else:
            uri = "mongodb://{host}:{port}".format(
                host=kwargs.get('host', 'localhost'),
                port=kwargs.get('port', 27017)
            )
        client = motor_asyncio.AsyncIOMotorClient(uri)
        return cls(client=client)

    async def find_data_base(self, name):
        data_bases = await self.client.list_database_names()
        if name in data_bases:
            return self.client[name]
        return None

    @async_action
    async def update(self, db, table, data, create=True, **kwargs):
        if data.get('_id'):
            res = await self.client[db][table].update_one(dict(_id=data['_id']), {'$set': data}, upsert=create)
        else:
            if create:
                res = await self.client[db][table].insert_one(data)
                data['_id'] = res.inserted_id
            else:
                raise KeyError
        return res

    @async_action
    async def push(self, db, table, uid, field, item, **kwargs):
        res = await self.client[db][table].update_one({'_id': uid}, {'$push': {field: item}}, upsert=False)
        return res

    @async_action
    async def pull(self, db, table, uid, field, item, **kwargs):
        kwargs.pop('_action')
        res = await self.client[db][table].update_one({'_id': uid}, {'$pull': {field: item}}, upsert=False)
        return res

    @async_action
    async def find_one(self, db, table, _filter, **kwargs):
        kwargs.pop('_action')
        return await self.client[db][table].find_one(_filter, **kwargs)

    async def delete_one(self, db, table, _filter):
        return await self.client[db][table].delete_one(_filter)

    async def delete_many(self, db, table, _filter):
        return await self.client[db][table].delete_many(_filter)

    async def count(self, db, table, **kwargs):
        return await self.client[db][table].count_documents(
            kwargs.get('filter', {})
        )

    @async_action
    async def query(self, db, table, **kwargs):
        cursor = self.client[db][table].find(
            filter=kwargs.get('filter', None),
            projection=kwargs.get('projection', None),
            skip=kwargs.get('skip', 0),
            limit=kwargs.get('limit', 1000)
        )
        sort = kwargs.get('sort')
        if sort:
            cursor.sort(sort)
        result = await cursor.to_list(length=1000)
        await cursor.close()
        return result

    @async_action
    async def pipeline(self, db, table, pipeline, **kwargs):
        projection = kwargs.get('projection')
        filter = kwargs.get('filter')
        skip = kwargs.get('skip', 0)
        sort = kwargs.get('sort')
        limit = kwargs.get('limit', 1000)
        _pipeline = []
        _pipeline += pipeline
        if filter:
            _pipeline.append({'$match': filter})
        if projection:
            _pipeline.append({'$project': projection})
        if sort:
            _pipeline.append({'$sort': sort})
        if skip:
            _pipeline.append({'$skip': skip})
        if limit:
            _pipeline.append({'$limit': limit})

        cursor = self.client[db][table].aggregate(_pipeline)
        result = await cursor.to_list(length=1000)
        return result


    async def find_one_and_update(self, db, table, filter, data):
        return await self.client[db][table].find_one_and_update(filter, data)



import sqlite3
import os


class SqlLite:
    def __init__(self, **kwargs):
        self.path = kwargs.get('path')
        # self.client = kwargs.get('client')

    pass

    @classmethod
    def connect(cls, device, **kwargs):
        self = cls()
        return self

    async def find_data_base(self, name):
        data_bases = await self.client.list_database_names()
        if name in data_bases:
            return self.client[name]
        return None

    async def update(self, db, table, data, create=True):
        if data.get('_id'):
            res = await self.client[db][table].update_one(dict(_id=data['_id']), {'$set': data}, upsert=create)
        else:
            if create:
                res = await self.client[db][table].insert_one(data)
                data['_id'] = res.inserted_id
            else:
                raise KeyError
        return res

    async def find_one(self, db, table, filter):
        return await self.client[db][table].find_one(filter)

    async def delete_one(self, db, table, filter):
        return await self.client[db][table].delete_one(filter)

    async def delete_many(self, db, table, filter):
        return await self.client[db][table].delete_many(filter)

    async def count(self, db, table, **kwargs):
        return await self.client[db][table].count_documents(
            kwargs.get('filter', {})
        )

    def get_db(self, name, create=True):
        path = os.path.normpath(f'{self.path}/{name}.db')
        if not os.path.isfile(not create and not path):
            raise Exception('db not found')
        return sqlite3.connect(path)

    def get_table(self, name):
        pass

    async def query(self, db, table, **kwargs):
        db = self.get_db(db)
        cursor = self.client[db][table].find(
            filter=kwargs.get('filter', None),
            projection=kwargs.get('projection', None),
            skip=kwargs.get('skip', 0),
            limit=kwargs.get('limit', 0)
        )
        sort = kwargs.get('sort')
        if sort:
            cursor.sort(sort)
        result = await cursor.to_list(length=1000)
        await cursor.close()
        return result

    async def find_one_and_update(self, db, table, filter, data):
        return await self.client[db][table].find_one_and_update(filter, data)



from bson import ObjectId, DBRef
from Bubot.Helpers.Helper import Helper
from Bubot.Core.BubotHelper import BubotHelper
from Bubot.Helpers.ExtException import ExtException, KeyNotFound, HandlerNotFoundError
from Bubot.Helpers.Action import async_action
from Bubot.Core.ObjForm import ObjForm
from Bubot.Core.ObjModel import ObjModel


class Obj:
    file = __file__
    extension = False
    name = None

    def __init__(self, storage, **kwargs):
        self.storage = storage
        self.account_id = kwargs.get('account_id')
        self.lang = kwargs.get('lang')
        self.form_id = kwargs.get('form')
        self.data = None
        data = kwargs.get('data')
        if data:
            self.init_by_data(data)
        self.debug = False

    def init(self):
        self.data = dict(
            title=''
        )

    def init_by_data(self, data):
        self.init()
        if data:
            Helper.update_dict(self.data, data)

    @classmethod
    @async_action
    async def init_by_ref(cls, store, obj_link, **kwargs):
        try:
            _ref = obj_link['_ref']
        except KeyError as err:
            raise KeyNotFound(detail=err)
        obj_name = _ref.collection
        _id = _ref.id
        obj_class = BubotHelper.get_obj_class(obj_name)
        return obj_class(store, **kwargs)

    @async_action
    async def find_by_link(self, obj_link, **kwargs):
        return await self.find_by_id(obj_link['_ref'].id, **kwargs)

    @async_action
    async def find_by_id(self, _id, **kwargs):
        action = kwargs['_action']
        self.add_projection(kwargs)
        res = action.add_stat(
            await self.storage.find_one(self.db, self.name, dict(_id=_id), **kwargs))
        if res:
            self.init_by_data(res)
            return action.set_end(res)
        raise KeyNotFound()

    def get_link(self, properties=None):
        '''

        :param properties: список свойств объекта которые нужно включить в ссылку
        :return: объект ссылки
        '''

        result = {
            "_ref": DBRef(self.name, self.obj_id)
        }
        for elem in self.data:  # добаляем заголовок на всех языках
            if elem[:5] == 'title':
                result[elem] = self.data[elem]
        return result

    @property
    def obj_name(self):
        return self.name if self.name else self.__class__.__name__

    @property
    def obj_id(self):
        return self.data['_id']

    @obj_id.setter
    def obj_id(self, value):
        self.data['_id'] = value

    @property
    def db(self):
        return self.account_id

    @async_action
    async def query(self, **kwargs):
        return await self.storage.query(self.db, self.name, **kwargs)

    async def count(self, **kwargs):
        return await self.storage.count(self.db, self.name, **kwargs)

    @async_action
    async def update(self, data=None, **kwargs):
        return await self.storage.update(self.db, self.name, data if data else self.data, **kwargs)
        pass

    @async_action
    async def push(self, field, item):
        res = await self.storage.push(self.db, self.name, self.obj_id, field, item)
        return res

    @async_action
    async def pull(self, field, item):
        res = await self.storage.pull(self.db, self.name, self.obj_id, field, item)
        return res

    async def delete_one(self, _id=None, _filter=None):
        # _id = ObjectId(_id) if _id else self.obj_id
        _filter = _filter if _filter else dict(_id=_id)
        await self.storage.delete_one(self.db, self.name, _filter)
        pass

    async def delete_many(self, _filter):
        await self.storage.delete_many(self.db, self.name, _filter)
        pass

    async def create(self, data=None):
        await self.storage.createupdate(self.db, self.name, self.data)
        pass

    @classmethod
    def get_form(cls, form_name):
        return ObjForm.get_form(cls, form_name)

    def add_projection(self, dest_obj):
        if self.form_id:
            return ObjForm.add_projection(self, self.form_id, dest_obj)

    # @classmethod
    # def get_obj_type(cls):
    #     if cls._obj_type is None:
    #         from os import sep
    #         _path = cls.file.split(sep)
    #         if _path[-2] == cls.get_obj_name():
    #             cls._obj_type = _path[-3]
    #     return cls._obj_type
    # 
    # @classmethod
    # def get_obj_name(cls):
    #     return cls.name if cls.name else cls.__name__

    @classmethod
    def get_model(cls):
        if cls.model is None:
            cls.model = ObjModel.get(cls)
        return cls.model

    # @classmethod
    # def get_obj_table(cls):
    #     if cls._obj_table is None:
    #         cls._obj_table = f'{cls._obj_table_prefix}{cls.get_obj_name()}'
    #     return cls._obj_table

    async def find_by_keys(self, keys):
        for key in keys:
            try:
                return await self.find_by_key(key['name'], key['value'])
            except KeyError:
                pass
        raise KeyError

    async def find_by_key(self, key_name, key_value):
        res = await self.storage.find_one(self.db, self.name, dict(
            keys=dict(name=key_name, value=key_value)
        ))
        if res:
            self.init_by_data(res)
            return res
        raise KeyError
        pass

    @property
    def title(self, lang=None):
        return self.data['title']



from os.path import dirname, normpath, isdir, isfile
from json import load


class ObjForm:
    def __init__(self, **kwargs):
        self.data = kwargs.get('data')

    @classmethod
    def get_form(cls, obj_handler, form_name):
        form_path = normpath(f'{dirname(obj_handler.file)}/form/{form_name}.form.json')
        if not isfile(form_path):
            if not obj_handler.extension:
                return None
            for elem in obj_handler.__bases__:
                if hasattr(elem, 'get_form'):
                    return cls.get_form(elem, form_name)

        with open(form_path, 'r', encoding='utf-8') as file:
            return cls(data=load(file))

    def get_projection(self):
        fields = self.data.get('fields')
        if not fields:
            return {'_id': 1, 'title': 1}
        res = {}
        for elem in fields:
            if fields[elem]:
                res[elem] = True
        return res

    @classmethod
    def add_projection(cls, obj_handler, form_name, obj):
        self = cls.get_form(obj_handler.__class__, form_name)
        obj['projection'] = self.get_projection()




# from aiocoap import Message, NON, Code
from Bubot.Helpers.Coap.coap import Message, NON, Code
from Bubot.Core.DeviceLink import ResourceLink
from Bubot.Helpers.ExtException import ExtException, dumps_error
import urllib.parse
import cbor2


class OcfMessage:
    def __init__(self, **kwargs):
        self.fr = ResourceLink.init(kwargs.get('fr'))
        self.to = ResourceLink.init(kwargs.get('to'))
        self.code = kwargs.get('code')
        self.ri = kwargs.get('ri', {})  # Request Identifier
        self.cn = kwargs.get('cn', {})  # Content
        # self.uri_path = kwargs.get('uri_path')
        self.query = kwargs.get('query', {})
        self.token = kwargs.get('token', b'')
        self.mid = kwargs.get('mid', 0)
        # self.data = kwargs.get('data', {})
        self.raw_msg = None
        # self.code = kwargs.get('code', NON)
        pass

    @staticmethod
    def encode_query(query):
        result = []
        for key in query:
            if not isinstance(query[key], list):
                raise Exception('value query param must be only list')
            for elem in query[key]:
                result.append('{}={}'.format(key, elem))
        return result

    @staticmethod
    def decode_query(query):
        result = {}
        if not query:
            return None
        for elem in query:
            key, value = elem.split('=')
            if key not in query:
                result[key] = []
            result[key].append(value)
            pass
        return result


class OcfRequest(OcfMessage):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.obs = kwargs.get('obs')
        self.multicast = kwargs.get('multicast', False)
        self.op = kwargs.get('op')
        if not self.code or not self.op:
            if not self.code and self.op:
                self.code = self.map_crudn_to_coap_code(self.op)
            else:
                raise Exception('не заполнены обязательные параметры')

        # self.observe = None

    def encode_to_coap(self):
        params = dict(
            mtype=NON,
            mid=self.mid,
            code=self.code,
            token=self.token,
            uri_host=self.to.anchor,
            uri_query=self.encode_query(self.query),
            uri_path=self.to.href[1:].split('/'),
            content_format=60,
            accept=60,
            observe=self.obs,
            payload=cbor2.dumps(self.cn) if self.cn else b''
        )
        msg2 = Message(**params)

        try:
            parsed = urllib.parse.urlparse(self.to.get_endpoint(), allow_fragments=False)
            address = (parsed.hostname, parsed.port)
        except KeyError:
            address = None
        return msg2, address

    @classmethod
    def decode_from_coap(cls, msg, multicast=False):
        # message = Message.decode(raw_data, remote)
        self = cls(
            fr=ResourceLink.init_from_uri('coap://[{0}]:{1}'.format(msg.remote[0], msg.remote[1])),  # message.opt.uri_path, message.opt.uri_port))]
            to=ResourceLink.init_from_msg(msg),
            code=int(msg.code),
            op=cls.map_coap_code_to_crudn(msg.code.name),
            obs=msg.opt.observe,
            query=cls.decode_query(msg.opt.uri_query),
            token=msg.token,
            mid=msg.mid,
            multicast=multicast
        )
        self.raw_msg = msg
        self.operation = msg.code.name.lower()

        if msg.payload:
            self.cn = cbor2.loads(msg.payload)

        return self

    @staticmethod
    def map_coap_code_to_crudn(code):
        map_coap_to_crudn = {
            'post': 'update',
            'put': 'create',
            'delete': 'delete',
            'get': 'retrieve'
        }
        try:
            return map_coap_to_crudn[code.lower()]
        except KeyError:
            raise Exception('Unknown CRUDN operation ({0})'.format(code))

    @staticmethod
    def map_crudn_to_coap_code(operation):
        #    +------+--------+-----------+
        #    | Code | Name   | Reference |
        #    | 0.01 | GET    | [RFC7252] |
        #    | 0.02 | POST   | [RFC7252] |
        #    | 0.03 | PUT    | [RFC7252] |
        #    | 0.04 | DELETE | [RFC7252] |
        #    +------+--------+-----------+
        map_crudn_to_coap = {
            'create': 3,
            'retrieve': 1,
            'update': 2,
            'delete': 4,
        }
        return map_crudn_to_coap[operation.lower()]


class OcfResponse(OcfMessage):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # self.error = self.is_successful(kwargs['rs'].)
        self.rs = kwargs['rs']  # Response Code RFC 7252
        self.obs = kwargs.get('obs')  # Observe
        # self.observe = kwargs.get('observe', None)

        # +------+------------------------------+-----------+
        # | Code | Description                  | Reference |
        # +------+------------------------------+-----------+
        # | 2.01 | Created                      | [RFC7252] |
        # | 2.02 | Deleted                      | [RFC7252] |
        # | 2.03 | Valid                        | [RFC7252] |
        # | 2.04 | Changed                      | [RFC7252] |
        # | 2.05 | Content                      | [RFC7252] |
        # | 4.00 | Bad Request                  | [RFC7252] |
        # | 4.01 | Unauthorized                 | [RFC7252] |
        # | 4.02 | Bad Option                   | [RFC7252] |
        # | 4.03 | Forbidden                    | [RFC7252] |
        # | 4.04 | Not Found                    | [RFC7252] |
        # | 4.05 | Method Not Allowed           | [RFC7252] |
        # | 4.06 | Not Acceptable               | [RFC7252] |
        # | 4.12 | Precondition Failed          | [RFC7252] |
        # | 4.13 | Request Entity Too Large     | [RFC7252] |
        # | 4.15 | Unsupported Content - Format | [RFC7252] |
        # | 5.00 | Internal Server Error        | [RFC7252] |
        # | 5.01 | Not Implemented              | [RFC7252] |
        # | 5.02 | Bad Gateway                  | [RFC7252] |
        # | 5.03 | Service Unavailable          | [RFC7252] |
        # | 5.04 | Gateway ExtTimeoutError              | [RFC7252] |
        # | 5.05 | Proxying Not Supported       | [RFC7252] |
        # +------+------------------------------+-----------+

    def is_successful(self):
        """True if the code is in the successful subrange of the response code range"""
        return True if (64 <= self.code < 96) else False

    @classmethod
    def generate_error(cls, err, req, **kwargs):
        if isinstance(err, ExtException):
            _err = err.dumps()
        else:
            _err = dumps_error(err)
        self = cls(
            to=req.to,
            fr=req.fr,
            uri_query=req.query,
            token=req.token,
            mid=req.mid,
            rs=Code.UNPROCESSABLE_ENTITY,
            cn=_err
        )
        return self

    @classmethod
    def generate_answer(cls, data, req, **kwargs):
        self = cls(
            to=req.to,
            fr=req.fr,
            mid=kwargs.get('mid', req.mid),
            rs=kwargs.get('code', Code.CONTENT),
            cn=data,
            uri_query=req.query,
            token=req.token,
            obs=req.obs
        )
        # if req.raw_msg:
        #     self.raw_msg = req.raw_msg.copy()
        # self.raw_msg.code = Code.CONTENT
        return self

    def encode_to_coap(self):
        parsed = urllib.parse.urlparse(self.fr.get_endpoint(), allow_fragments=False)
        msg2 = Message(
            mtype=NON,
            mid=self.mid,
            code=self.rs,
            observe=self.obs,
            payload=cbor2.dumps(self.cn) if self.cn else b'',
            token=self.token,
            uri_query=self.encode_query(self.query),
            uri_host=self.to.anchor,
            uri_path=self.to.href[1:].split('/'),
            content_format=60,
            accept=60
            # iotivity_addresses=b'\xc0'
        )
        # message = self.raw_msg
        # message.code = self.code
        # message.payload = cbor2.dumps(self.data) if self.data else b''
        # message.mtype = NON
        return msg2, (parsed.hostname, parsed.port)

    @classmethod
    def decode_from_coap(cls, msg, multicast=False):
        # message = Message.decode(raw_data, remote)
        self = cls(
            code=int(msg.code),
            rs=msg.code.name.lower(),
            token=msg.token,
            mid=msg.mid,
            query=cls.decode_query(msg.opt.uri_query),
            obs=msg.opt.observe,
            multicast=multicast,
            to=ResourceLink.init_from_msg(msg)
        )
        self.raw_msg = msg
        self.operation = msg.code.name.lower()
        if msg.payload:
            self.cn = cbor2.loads(msg.payload)
        if msg.opt.uri_query:
            for elem in msg.opt.uri_query:
                key, value = elem.split('=')
                if key not in self.uri_query:
                    self.uri_query[key] = []
                self.uri_query[key].append(value)
                pass
        return self

    def to_dict(self):
        return dict(
            code=self.code,
            token=self.token,
            mid=self.mid,
            obs=self.obs,
            fr=self.fr.data if self.fr else None,
            to=self.to.data if self.to else None,
            cn=self.cn
        )

    def __str__(self):
        return '{code} {value}'.format(code=self.raw_msg.code, value=str(self.cn))



import argparse
from BubotObj.OcfDevice.subtype.Device.Device import Device
from Bubot.Helpers.ExtException import HandlerNotFoundError
import logging


def main():
    parser = argparse.ArgumentParser(description='IoT framework based on OCF specification')
    parser.add_argument('--n', dest='class_name', default='WebServer', type=str,
                        help='OCF device class name (default: WebServer)')
    parser.add_argument('--i', dest='di', type=str, help='ID of the device being launched, optional')
    parser.add_argument('--p', dest='path', default='./', type=str, help='path to config dir (default: ./)')
    # parser.add_argument('--l', default='./', type=str, help='path to log dir (default: ./)')
    parser.add_argument('--d', dest='log_level', default='error', type=str,
                        help='log level: notset/debug/info/warning/error/critical (default: error)')
    # parser.add_argument('--f', dest='log_format', default='%(levelname)s %(name)s.%(funcName)s %(message)s',
    #                     type=str, help=f'logger format (default: ')
    args = parser.parse_args()

    logging.basicConfig(
        level=args.log_level.upper(),
        format='%(levelname)s %(name)s.%(funcName)s %(message)s'
    )
    try:
        device = Device.init_from_file(di=args.di, path=args.path, class_name=args.class_name)
    except HandlerNotFoundError:
        print(f'Error: class_name {args.class_name} not found. run "bubot -h" for help')
        return
    device.run()



# import sys
import logging
from multiprocessing import Lock
import multiprocessing



class Logger:
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0
    _levels = {
        0: 'NOTSET',
        10: 'DEBUG',
        20: 'INFO',
        30: 'WARNING',
        40: 'ERROR',
        50: 'CRITICAL'
    }

    def __init__(self, device, **kwargs):
        log = kwargs.get('log')
        if log:
            self.lock = log.lock
        else:
            self.lock = Lock()
        self._level = None
        self._level_name = None
        self.device = device
        self.level = kwargs.get('level', 'error')
        self.format = kwargs.get(
            'format',
            '%(levelname)s %(name)s %(funcName)s %(message)s'
        )
        # if not device or not log or isinstance(self.device, multiprocessing.Process):
        #     self.logger = multiprocessing.get_logger()
        # else:
        #     self.logger = log.logger
        self.name = kwargs.get('name', device.__class__.__name__)

    def debug(self, msg, *args, **kwargs):
        self.log(10, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.log(20, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.log(30, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.log(40, msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.log(50, msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        if level >= self._level:
            level_name = self._levels[level]
            self.lock.acquire()
            try:
                msg = self.format % dict(levelname=level_name, name=self.name, message=msg, funcName='')
                print(msg)
                # self.logger.log(level, message)
            finally:
                self.lock.release()

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, value):
        self._level = getattr(self, value.upper())
        self._level_name = value.upper()



from BubotObj.OcfDevice.subtype.Device.Device import Device
from datetime import datetime
from .__init__ import __version__ as device_version
import asyncio


# _logger = logging.getLogger(__name__)


class Console(Device):
    version = device_version
    file = __file__
    template = False

    async def on_idle(self):
        i = datetime.now().strftime('%d.%m.%Y %H:%M:%S')
        self.set_param('/oic/mnt', 'DateTime', i)
        self.log.info('DateTime: {}'.format(i))
        await asyncio.sleep(1)

    def get_install_actions(self):
        result = [
            dict(
                id='search_devices',
                icon='mdi-radar',
                title='search OcfDriver'
            )
        ]
        result.extend(super().get_install_actions())
        return result

    async def on_action(self, message, answer):
        j = self.get_param('/oic/mnt', 'j')
        self.set_param('/oic/mnt', 'j', j + 1)
        self.log.info('j: {}'.format(j))

    async def on_update_console(self, message, answer):
        self.log.debug('{0} receive notify {1} {2}'.format(
            self.__class__.__name__, message.to.di, message.to.href))
        self.log.info(str(message))



from py4j.java_gateway import JavaGateway, CallbackServerParameters, GatewayParameters
import json
import uuid


class JasperReport:
    data_dir = "./jasper_reports"
    address = '192.168.1.28'

    def pdf_from_json(self, report, data):
        gateway = JavaGateway(
            gateway_parameters=GatewayParameters(address=self.address),
        )
        uid = str(uuid.uuid4())
        with open('{0}/request/{1}.json'.format(self.data_dir, uid), 'w') as file:
            json.dump(data, file)
        gateway.entry_point.pdf_from_json(report, uid)
        return uid



from aiohttp import web
from Bubot.Helpers.Helper import Helper


class ReportHandler(web.View):
    def __init__(self, request):
        web.View.__init__(self, request)
        self.obj_type = self.request.match_info.get('objType')
        self.obj_name = self.request.match_info.get('objName')
        self.report_name = self.request.match_info.get('reportName')
        self.report_section = self.request.match_info.get('reportSection')
        try:
            self.handler = Helper.get_obj_class(f'jay.{self.obj_type}.{self.obj_name}', 'reports', self.report_name)()
        except Exception as err:
            raise err

    async def get(self):
        try:
            handler = getattr(self.handler, self.report_section)
            return await handler(self)
        except Exception as err:
            return web.HTTPInternalServerError(text=str(err))



import urllib.parse
import asyncio
from Bubot.Helpers.Coap.coap import Message
from Bubot.Helpers.Helper import ArrayHelper
from Bubot.Helpers.ExtException import ExtException


# _logger = logging.getLogger(__name__)


class ResourceLink:
    def __init__(self):
        self.data = {}
        self.bm = 0
        self.observe = False

    @classmethod
    def init(cls, value):
        if isinstance(value, str):  # uri
            return cls.init_from_uri(value)
        if isinstance(value, dict):
            return cls.init_from_link(value)
        if isinstance(value, Message):
            return cls.init_from_msg(value)
        if isinstance(value, ResourceLink):
            return value
        return None

    @classmethod
    def init_from_uri(cls, uri):
        self = cls()
        self.parse_uri(uri)
        return self

    @classmethod
    def init_from_msg(cls, msg):
        self = cls()
        self.parse_uri(msg.opt.uri_host)
        self.parse_uri('coap://{0}:{1}'.format(msg.remote[0], msg.remote[1]))
        self.href = '/{}'.format('/'.join(msg.opt.uri_path))
        return self

    @classmethod
    def init_from_link(cls, link, **kwargs):
        self = cls()
        self._parse_link(link)
        return self

    def set_data(self, _name, src, *args):
        try:
            self.data[_name] = src[_name]
        except KeyError:
            try:
                self.data[_name] = args[0]
            except IndexError:
                pass

    @property
    def uid(self):
        if self.di:
            uid = 'ocf://{}'.format(self.di)
        else:
            uid = self.get_endpoint()

        href = self.href
        if href:
            uid += href
        return uid

    @classmethod
    def init_from_device(cls, device, href, **kwargs):
        new_data = device.data[href]
        self = cls()
        self.di = device.di
        self.data['anchor'] = 'ocf://{}'.format(self.di)
        self.data['eps'] = device.eps
        self.data['href'] = href
        self.set_data('id', new_data)
        self.set_data('rt', new_data)
        self.set_data('if', new_data)
        self.set_data('n', new_data)
        self.set_data('p', new_data, dict(bm=0))
        return self

    def get(self, name, default=None):
        return self.data.get(name, default)

    @staticmethod
    def get_default_data():
        return {
            # 'di': '',  # device id
            'href': '',  # resource
            'anchor': '',  # ocf uri
            'title': '',  # ocf uri
            'id': '',  # id resource
            'n': '',  # name resource
            'rt': [],
            'if': [],
            'eps': [],
            'p': {}
        }

    def _parse_link(self, link):
        self.data = self.get_default_data()
        self.set_data('di', link)
        if self.di:
            self.data['anchor'] = 'ocf://{}'.format(self.di)
        else:
            self.parse_uri(link.get('anchor'))
        self.parse_uri(link.get('href'))
        self.set_data('id', link)
        self.set_data('rt', link)
        self.set_data('if', link)
        self.set_data('n', link)
        self.set_data('p', link, dict(bm=0))
        if 'eps' in link:
            for ep in link['eps']:
                ArrayHelper.update(self.data['eps'], ep, 'ep')
        if 'ep' in link:
            ArrayHelper.update(self.data['eps'], dict(ep=link['ep']), 'ep')

    def parse_uri(self, _uri):
        if _uri is None:
            return
        uri = urllib.parse.urlparse(_uri)  # (scheme, netloc, path, params, query, fragment)
        if uri[0]:
            if uri[0] == 'ocf':
                self.data['di'] = uri[1]
            elif uri[0] == 'coap' or uri[0] == 'coaps':
                if 'eps' not in self.data:
                    self.data['eps'] = []
                self.data['eps'].append(dict(ep='{0}://{1}'.format(uri[0], uri[1])))
        if uri[2]:
            self.data['href'] = uri[2]

    async def retrieve(self, sender_device):
        # _log.debug('retrieve {}'.format(self.href))
        _data = await sender_device.request(
            'retrieve',
            self.data,
            None
        )
        # _log.debug('retrieve {} {}'.format(self.href, _data.cn))
        return _data.cn

    @property
    def di(self):
        return self.data.get('di')

    @di.setter
    def di(self, value):
        self.data['di'] = value

    @property
    def href(self):
        return self.data.get('href')

    @href.setter
    def href(self, value):
        self.data['href'] = value

    @property
    def anchor(self):
        if self.data.get('anchor'):
            return self.data['anchor']
        if self.data.get('di'):
            return 'ocf://{}'.format(self.data['di'])

    @property
    def name(self):
        if self.data.get('n'):
            return self.data['n']
        return self.data['href']

    def get_endpoint(self):
        try:
            return self.data['eps'][0]['ep']
        except IndexError:
            return ''

    @property
    def discoverable(self):
        _bm = self.data['p'].get('bm', 0)
        if _bm > 0:
            return int(bin(_bm)[-1])
        return 0


class DeviceLink:
    def __init__(self):
        self.links = {}
        self.data = {}
        self.di = None
        # self.n = None
        self.eps = None

    @classmethod
    def get_default_link(cls):
        return {
            'di': '',  # device id
            'href': '',  # resource
            'anchor': '',  # ocf uri
            'rt': [],
            'if': [],
            'eps': [],
            'p': {}
        }

    @classmethod
    def get_resource_by_uri(cls, resources, uri):
        link = ResourceLink.init_from_uri(uri)
        if not link.di:
            raise ExtException('bad schema id, need ocf')
        try:
            return resources[link.di].links[link.href]
        except KeyError:
            raise ExtException('resource not found', detail=uri)

    @classmethod
    def init_from_oic_res(cls, data):
        self = cls()
        self.di = data['di']
        # self.n = data.get('n', '')
        for link in data['links']:
            data = ResourceLink.init_from_link(link, di=self.di)
            self.links[data.get('href')] = data
            self.eps = data.data['eps']
        return self

    def update_from_oic_res(self, data):
        raise NotImplemented()

    async def retrieve(self, sender_device, link):
        _data = self.links[link].retrieve(sender_device)
        return _data

    async def retrieve_all(self, sender_device):
        requests = []
        index = []
        result = []
        for href in self.links:
            if href == '/oic/res':
                continue
            index.append(href)
            requests.append(self.links[href].retrieve(sender_device))
        result = await asyncio.gather(*requests, return_exceptions=True, loop=sender_device.loop)
        for i in range(len(result)):
            if not isinstance(result[i], Exception):
                self.data[index[i]] = result[i]
        pass

    @classmethod
    def init_from_device(cls, device):
        self = cls()
        self.data = device.data
        self.di = device.get_device_id()
        self.eps = self.get_device_eps(device)
        for res in self.data:
            self.links[res] = ResourceLink.init_from_device(self, res)
        return self

    @staticmethod
    def get_device_eps(device):
        eps = []
        if device.coap and device.coap.endpoint:
            for elem in device.coap.endpoint:
                if elem == 'multicast' or not device.coap.endpoint[elem]:
                    continue
                eps.append(dict(ep=device.coap.endpoint[elem]['uri']))
        return eps

    @property
    def name(self):
        # if self.n:
        #     return self.n
        try:
            _name = self.links['/oic/con'].data['n']
            if _name:
                return _name
            # return self.n
        except KeyError:
            pass
        try:
            return self.links['/oic/d'].data['n']
            # return self.n
        except KeyError:
            return ''

    def get(self, param_path, *args):
        _param_path = param_path.split('.')
        _data = self.data
        for elem in _param_path:
            try:
                _data = _data[elem]
            except KeyError:
                if len(args) > 0:
                    return args[0]
                raise KeyError(param_path)
        return _data

    def to_object_data(self):
        res = []
        for href in self.links:
            data = self.links[href].data
            if 'rt' not in data or 'if' not in data:
                raise ExtException('bad resource', detail=f'{href} - rt or if not defined')
            res.append({
                'href': href,
                'rt': self.links[href].data['rt'],
                'if': self.links[href].data['if'],
                'n': self.links[href].name
            })
        return {
            '_id': self.di,
            'n': self.name,
            'ep': self.eps[0]['ep'] if self.eps else None,
            'res': res
        }



import setuptools
from src.BubotObj.OcfDevice.subtype.Device import __version__

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='Bubot_Core',
    version=__version__,
    author="Razgovorov Mikhail",
    author_email="1338833@gmail.com",
    description="iot framework based on OCF specification",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/businka/Bubot_Core",
    package_dir={'': 'src'},
    package_data={
        '': ['*.md', '*.json'],
    },
    packages=setuptools.find_namespace_packages(where='src'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Framework :: AsyncIO",
    ],
    python_requires='>=3.7',
    zip_safe=False,
    install_requires=[
        'cbor2>5',
        'Bubot_Helpers'
    ],
    entry_points={
        'console_scripts':
            [
                'bubot = Bubot.Core:main',
                'Bubot = Bubot.Core:main'
            ]
    }
)



import asyncio
from Bubot.Helpers.ExtException import ExtException


# _logger = logging.getLogger(__name__)


class QueueMixin:

    async def queue_worker(self, queue, name=''):
        self.log.debug('start queue_worker "{}"'.format(name))
        while True:
            (future, result) = await queue.get()
            try:
                _result = await future
                if result and asyncio.isfuture(result):
                    result.set_result(_result)
            except ExtException as e:
                result.set_exception(e)
            except Exception as e:
                result.set_exception(ExtException(parent=e, action='queue_worker'))
            finally:
                queue.task_done()

    async def execute_in_queue(self, queue, task, name=''):
        try:
            self.log.debug('execute_in_queue {}'.format(name))
            result = asyncio.Future()
            queue.put_nowait((task, result))
            return await result
        except ExtException as e:
            raise ExtException(parent=e) from None
        except Exception as e:
            raise ExtException(parent=e, action='QueueMixin.execute_in_queue')



from datetime import datetime
from Bubot.Core.Obj import Obj


class DocumentObj(Obj):
    name = None

    def init(self):
        self.data = {
            "title": "",
            "name": "",
            "date": datetime.now(),
            "number": "",
            "keys": []
        }

    # async def find_by_keys(self, keys):
    #     for key in keys:
    #         try:
    #             return await self.find_by_key(key['name'], key['value'])
    #         except KeyError:
    #             pass
    #     raise KeyError
    #
    # async def find_by_key(self, key_name, key_value):
    #     res = await self.db.find_one(self.db, self.name, dict(
    #         keys=dict(name=key_name, value=key_value)
    #     ))
    #     if res:
    #         self.init_by_data(res)
    #         return res
    #     raise KeyError
    #     pass

    @property
    def title(self, lang=None):
        return self.data['title']

    @property
    def obj_type(self):
        return 'Document'
